/* exithl.c
**
** REMOVABLE exit functions (based on a handler which is automatically
** installed with `atexit' when the first exit handler is installed).
*/

/* Ich bentige (teilweise) Aufrumfunktionen, die auch wieder entfernt werden
** knnen; dazu installiere ich einmal mittels `atexit' eine Funktion
** `exit_handler', die einfach alle registrierten Aufrumfunktionen in
** umgekehrter Reihenfolge ihrer Registrierung ausfhrt.
** Zum Registrieren verwende ich eine einfache lineare Liste. Diese
** bietet zum einen den Vorteil, da die Aufrumfunktionen in der fr die
** Ausfhrung richtigen Reihenfolge (umgekehrt zu ihrer Registrierung) gefunden
** werden; zum anderen lassen sich einzelne Aufrumfunktionen auch wieder aus
** dieser Liste entfernen (Aufhebung der Registrierung).
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vxerror.h"
#include "tools.h"
#include "mrmacs.h"

/* Mit der folgenden Variablen kann festgestellt werden, ob die Routine
** zur Ausfhrung der Aufrumfunktionen bereits installiert wurde oder nicht.
*/
static int exit_handler_installed = 0;

/* Aufrumfunktionen mssen - so oder so - in umgekehrter Reihenfolge ihrer
** Installation (Registrierung) ausgefhrt werden; darum reicht zur ihrer
** Verwaltung auch eine einfache lineare Liste, die LIFO-artig benutzt wird.
*/
static struct exit_list {
    struct exit_list *next;
    char *handle_name;
    void (*exit_func) (void);
} *handler_list = (struct exit_list *) NULL;

/* Dies ist der `exit handler'; diese Funktion wird einmal installiert und
** fhrt danach jede Aufrumfunktion, die in der Liste steht einmal auf, sobald
** sie aufgerufen wird (beim Verlassen des jeweiligen Programmes durch `exit').
*/
static void exit_handler (void)
{
    struct exit_list *c;
    for (c = handler_list; noNULL(c); c = c->next) {
	c->exit_func ();
    }
}

/* add_exithl (name, exit_func)
**                   - nimmt eine Aufrumfunktion unter dem Namen `name' in
**                     die Liste auf. Falls der `exit handler' (s.o.) noch
**                     nicht installiert wurde, dann wird er zuerst installiert.
*/
void add_exithl (char *handle_name, void (*exit_func) (void))
{
    struct exit_list *new;
    size_t newl = sizeof(struct exit_list) + strlen (handle_name) + 1;
    char *p;
    if (!exit_handler_installed) {
	atexit (exit_handler);
	exit_handler_installed = 1;
    }
    new = galloc (__FILE__, __LINE__, newl);
    new->next = handler_list;
    p = (char *) new + sizeof (struct exit_list);
    new->handle_name = p;
    strcpy (p, handle_name);
    new->exit_func = exit_func;
    handler_list = new;
}

/* del_exithl (name) - entfernt jede Aufrumfunktion aus der Liste, die dem
**                     Namen `name' zugeordnet wurde.
*/
void del_exithl (char *handle_name)
{
    struct exit_list *rmv, *prv;
    ifnull (handler_list) { return; }
    while (!strcmp (handler_list->handle_name, handle_name)) {
	rmv = handler_list;
	handler_list = rmv->next;
	gfree (rmv);
    }
    ifnull (handler_list) { return; }
    prv = handler_list;
    rmv = prv->next;
    while (noNULL(rmv)) {
	if (!strcmp (rmv->handle_name, handle_name)) {
	    prv->next = rmv->next;
	    gfree (rmv);
	    rmv = prv;
	}
	prv = rmv;
	rmv = rmv->next;
    }
}

/* clean_exithl ()   - entfernt alle Aufrumfunktionen aus der Liste.
*/
void clean_exithl (void)
{
    struct exit_list *curr;
    while (noNULL(handler_list)) {
	curr = handler_list;
	handler_list = curr->next;
	gfree (curr);
    }
}

/* is_exithl (name)  - liefert 1 (wahr), falls zu dem angegebenen Namen `name'
**                     eine Aufrumfunktion existiert oder 0 (falsch) sonst.
*/
int is_exithl (char *handle_name)
{
    struct exit_list *curr = handler_list;
    while (noNULL(curr)) {
	if (!strcmp (curr->handle_name, handle_name)) { return 1; }
	curr = curr->next;
    }
    return 0;
}

/* run_exithl (name) - fhrt die dem Namen `name' zugeordnete Aufrumfunktion
**                     aus oder (nacheinander) alle, falls NULL fr `name'
**                     bergeben wurde.
*/
void run_exithl (char *handle_name)
{
    struct exit_list *curr = handler_list;
/*    if (handle_name != (struct exit_list *) NULL) {*/
    unlessnull (handle_name) {
	while (noNULL(curr) && strcmp (curr->handle_name, handle_name)) {
	    curr = curr->next;
	}
	unlessnull (curr) { curr->exit_func (); }
    } else {
	while (noNULL(curr)) {
	    curr->exit_func ();
	    curr = curr->next;
	}
    }
}
